#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#include "ode-solver.h"
#include "macros.h"


/* 
 * ode_solver_modify_h - step correction method.
 *
 * sys:      pointer to ode solver system instance
 * epsilon:  tolerance
 * h:        pointer to step
 * y_err:    relative error
 *
 * Returns: ODE_SOLVER_H_INCREASE if step increases
 *          ODE_SOLVER_H_DECREASE if step decreases
 *          error if invalid parameters
 *          ODE_SOLVER_H_ACCEPT else
 */

int
ode_solver_modify_h(struct ode_solver_system *sys,
		    const float epsilon,
		    float *h, const float y_err[])
{
	float max_err, q;
	int i, n;

	return_if_fail(y_err != NULL);
	return_if_fail(h != NULL);
	return_if_fail(sys != NULL);

	max_err = 0.0;
	n       = sys->dimension;

	for (i = 0; i < n; i++)
		max_err = (max_err > fabs(y_err[i])) ? max_err : fabs(y_err[i]);

	if (max_err != 0.0)
		q = 0.84 * sqrt(sqrt(*h * epsilon / max_err));
	else
		q = 4.0;

	if (q >= 4.0) {
		*h *= q;
		return ODE_SOLVER_H_INCREASE;
	}

	if (q < 0.1) {
		*h *= q;
		return ODE_SOLVER_H_DECREASE;
	}

	return ODE_SOLVER_H_ACCEPT;
}


/* 
 * ode_solver_free - destroy ODE solver instance.
 *
 * solver: pointer to ode solver instance
 */

void
ode_solver_free(struct ode_solver *solver)
{
	free(solver);
}

/* 
 * Perform integration step from x to x+h using method Runge-Kutta 
 * of the fouth order.
 */

static int 
method_rk4(struct ode_solver *solver,
	   struct ode_solver_system *sys,
	   float x, float h, float y[],
	   const float dy_in[], float dy_out[],
	   float y_err[])
{
	float *k;
	float hh, h3, h6;
	int i, n;

	hh = 0.5 * h;
	h3 = h / 3.0;
	h6 = h / 6.0;

	n = sys->dimension;

	k = alloca(sizeof(float) * n);

	(*sys->model)(x, dy_in, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		y[i] = dy_in[i] + h6 * dy_out[i];
		k[i] = dy_in[i] + hh * dy_out[i];
	}
	
	(*sys->model)(x + hh, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		y[i] += h3 * dy_out[i];
		k[i] = dy_in[i] + hh * dy_out[i];
	}

	(*sys->model)(x + hh, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		y[i] += h3 * dy_out[i];
		k[i] = dy_in[i] + h * dy_out[i];
	}

	(*sys->model)(x + h, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++)
		y[i] += h6 * dy_out[i];

	return 0;
}

/* 
 * Perform integration step from x to x+h using method Runge-Kutta-Fehlberg. 
 */
static int
method_rkf45(struct ode_solver *solver,
	     struct ode_solver_system *sys,
	     float x, float h,
	     float y[],
	     const float dy_in[], float dy_out[],
	     float y_err[])
{
	float *k, *k1, *k2, *k3, *k4, *k5;
	int i, n;

	n  = sys->dimension;
	k1 = alloca(sizeof(float) * n);
	k2 = alloca(sizeof(float) * n);
	k3 = alloca(sizeof(float) * n);
	k4 = alloca(sizeof(float) * n);
	k5 = alloca(sizeof(float) * n);
	k  = alloca(sizeof(float) * n);

	(*sys->model)(x, dy_in, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		k1[i] = h * dy_out[i];
		k[i]  = dy_in[i] +  0.25 * k1[i];
	}

	(*sys->model)(x + 0.25 * h, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		k2[i] = h * dy_out[i];
		k[i]  = dy_in[i] + (3.0 / 32.0) * k1[i] + (9.0 /32.0) * k2[i];
	}

	(*sys->model)(x + 3.0 * h / 8.0, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		k3[i] = h * dy_out[i];
		k[i]  = dy_in[i] +
		       (1932.0 * k1[i] - 7200.0 * k2[i] +
		        7296.0 * k3[i]) / 2197.0;
	}

	(*sys->model)(x + 12.0 / 13.0 * h, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		k4[i] = h * dy_out[i];
		k[i] = dy_in[i] + (439.0 / 216.0) * k1[i] - 8.0 * k2[i] +
		       (3680.0 / 513.0) * k3[i] - (845.0 / 4104.0) * k4[i];
	}

	(*sys->model)(x + h, k, dy_out, sys->opaque);

	for (i = 0; i < n; i++) {
		k5[i] = h * dy_out[i];
		k[i] = dy_in[i] - (8.0 /27.0) * k1[i] + 2 * k2[i] -
		       (3544.0 / 2565.0) * k3[i] + (1859.0 / 4104.0) * k4[i] -
		       (11.0 / 40.0) * k5[i];
	}

	(*sys->model)(x + 0.5 * h, k, dy_out, sys->opaque);
	
	if (y_err != NULL)
		for (i = 0; i < n; i++) {
			y_err[i] = (2.0 / 55.0) * h * dy_out[i] +
				   (1.0 / 360.0) * k1[i] - (128.0 / 4275.0) * k3[i] -
				   (2282.683 / 77139.81) * k4[i] + 0.02 * k5[i];
		}

	for (i = 0; i < n; i++) {
		y[i] = dy_in[i] + (25.0 / 216.0) * k1[i] +
		       (1408.0 / 2565.0) * k3[i] +
		       (2197.0 / 4101.0) * k4[i] - 0.2 * k5[i];
	}

	return 0;
}

/* 
 * Perform integration step from x to x+h using solver `solver' and system `sys'.
 *
 * solver:  pointer to ode solver instance
 * sys:     pointer to ode solver system instance
 * x:       initial time
 * h:       integration step
 * y:       resulting array
 * dy_in:   initial conditions array
 * dy_out:  array of derivatives
 * y_err:   relative error
 *
 * Returns: error if invalid parameters
 *          ODE_SOLVER_OK else
 */
 
int
ode_solver_step(struct ode_solver *solver,
		struct ode_solver_system *sys,
		float x, float h,
		float y[],
		const float dy_in[], float dy_out[],
		float y_err[])
{
	int res;

	return_if_fail(solver != NULL);
	return_if_fail(sys != NULL);
	return_if_fail(dy_in != NULL);
	return_if_fail(dy_out != NULL);
	return_if_fail(y != NULL);
	return_if_fail(sys->model != NULL);

	res = (*solver->solver_step)(solver, sys, x, h, y, dy_in, dy_out, y_err);

	return res;
}

/*
 * Create RK4 ODE solver.
 *
 * Returns: ode_solver structure or NULL if invalid parameters.
 */

struct ode_solver *
ode_solver_rk4_new(void)
{
	struct ode_solver *ptr;

	ptr = malloc(sizeof(struct ode_solver));

	if (ptr == NULL)
		return NULL;

	memset(ptr, 0, sizeof(*ptr));

	ptr->name           = "rk4";
	ptr->evaluates_yerr = FALSE;
	ptr->solver_step    = method_rk4;
	ptr->solver_deinit  = ode_solver_free;

	return ptr;
}

/* 
 * Create RKF45 ODE solver.
 *
 * Returns: ode_solver structure or NULL if invalid parameters.
 */

struct ode_solver *
ode_solver_rkf45_new(void)
{
	struct ode_solver *ptr;

	ptr = malloc(sizeof(struct ode_solver));

	if (ptr == NULL)
		return NULL;

	memset(ptr, 0, sizeof(*ptr));

	ptr->name           = "rkf45";
	ptr->evaluates_yerr = TRUE;
	ptr->solver_step    = method_rkf45;
	ptr->solver_deinit  = ode_solver_free;

	return ptr;
}

